/*
 * Copyright (C) 1998  Warren Jasper
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 *
 */

/***************************************************************************
 *
 * a2dc.c
 *
 ***************************************************************************/

#define MODULE

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <stdlib.h>
#include "das08.h"
#include "a2dc.h"

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

static int  SoftRead(WORD *outBuff, ADC_ChanRec *chan);
static int  TrigRead(WORD *outBuff, ADC_ChanRec *chan);
static void SetGains(BYTE gain);
static void LoadCounter0( LONG value );
static void LoadCounter1( LONG value );
static void LoadCounter2( LONG value );

static int das08_read(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int das08_write(struct inode *iNode, struct file *filePtr, const char *buf, int count);
static int das08_open(struct inode *iNode, struct file *filePtr);
static void das08_close(struct inode *iNode, struct file *filePtr);
static int das08_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);
static void das08_ReadInterrupt();
static void das08_TimerHandler(LONG);
int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static WORD base_reg =  DAS08_BASE;         /* base register address */
static int WordsToRead;                     /* number of conversions until done  */
static BYTE dio_reg = 0x80;                 /* DIO control register              */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in          */
static ADC_ChanRec *CurrChan = NULL;        /* pointer to ChanADC[minor]         */
static ADC_ChanRec ChanADC[AD_CHANNELS];    /* ADC Channel specific information  */
static DAC_ChanRec ChanDAC[DA_CHANNELS];    /* DAC Channel specific information  */
static DIO_ChanRec ChanDIO[DIO_PORTS];      /* DIO Channel specific information  */
static BoardRec BoardData;                  /* Board specific information        */
static WORD KernBuff[MAX_COUNT];            /* Kernel buffer where samples are   */
                                            /* saved before write to user space  */
static WORD* KernBuffPtr;                   /* pointer to kernbuf                */
#if LINUX_VERSION_CODE < 0x020000
char kernel_version[] = UTS_RELEASE;        /* Required to by module             */
#endif
struct wait_queue  *adc0_wait;              /* wait semaphore                    */
struct timer_list TimerList = {NULL, NULL, 0, 0, das08_TimerHandler};


/***************************************************************************
 *
 *
 ***************************************************************************/

static struct file_operations das08_fops = {
	NULL,           /* das08_seek */
	das08_read,
	das08_write,    /* das08_write */
	NULL,		/* das08_readdir */
	NULL,		/* das08_select */
	das08_ioctl,	/* das08_ioctl */
	NULL,
	das08_open,
	das08_close,
	NULL		/* fsync */
};

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************
 *
 * Loads driver. Called when "insmod das08.o" is invoked on the command line.
 *               The board is set to IRQ, 
 *
 ***************************************************************************/

int init_module(void) 
{
  int err;
  int i;

  /* Register as a device with kernel.  */

  if ((err = register_chrdev(MajorNumber, "das08-AOL", &das08_fops))) {
      printk("%s: Failure to load module. error %d\n", ADAPTER_ID, -err);
      return err;
  } 

  /* Register the region of iospace with kernel and get base address */

  if ( check_region(base_reg, BOARD_SIZE ) != 0 ) {  /* in use */
    printk("%s: Can't allocate region port address %#x\n", 
            ADAPTER_ID, base_reg );
    if (unregister_chrdev(MajorNumber, "das08-AOL") != 0) {
      printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
    }
    return -ENODEV;
  } else {
    request_region( base_reg, BOARD_SIZE, "das08-AOL" );
  }
  BoardData.base = base_reg;

  /* Register interrupt handler. */

  BoardData.irq = DAS08_IRQ;

#if LINUX_VERSION_CODE > 0x020000
  if (request_irq(BoardData.irq, das08_ReadInterrupt, SA_INTERRUPT, "das08-AOL", NULL)) {
#else
  if (request_irq(BoardData.irq, das08_ReadInterrupt, SA_INTERRUPT, "das08-AOL")) {
#endif
    printk("%s: Can't request IRQ %d\n", ADAPTER_ID, BoardData.irq);
    if ( unregister_chrdev( MajorNumber, "das08-AOL" ) != 0 ) {  /* unregister device */
        printk( "%s: unregister_chrdev() failed.\n", ADAPTER_ID);
    }
    release_region( BoardData.base, BOARD_SIZE );       /* release region */
    return -(EIO);
  }
  /* Set all channel structures to show nothing active/open */

  for (i = 0; i < AD_CHANNELS; i++) {
      ChanADC[i].open = FALSE;
      ChanADC[i].gain = UP_10_00V;   /* set range to 0-10V */
      ChanADC[i].chan = i;           /* set channel number */

  }

  adc0_wait = NULL;

  for ( i = 0; i < DA_CHANNELS; i++ ) {
      ChanDAC[i].open = FALSE;
  }

  ChanDAC[0].addr = DA_CHAN0_LSB;
  ChanDAC[0].gain = DAC_GAIN0;
  ChanDAC[1].addr = DA_CHAN1_LSB;
  ChanDAC[1].gain = DAC_GAIN1;

  /* zero the output */
  outb(0x0, DA_CHAN0_LSB);
  outb(0x0, DA_CHAN0_MSB);
  outb(0x0, DA_CHAN1_LSB);
  outb(0x0, DA_CHAN1_MSB);

  ChanDIO[0].addr = DIO_PORTA;
  ChanDIO[1].addr = DIO_PORTB;
  ChanDIO[2].addr = DIO_PORTC;

  for ( i = 0; i < DIO_PORTS; i++ ) {
    ChanDIO[i].open = FALSE; 
    ChanDIO[i].mode = 0;
    outb_p(0x0, ChanDIO[i].addr);
  }


  BoardData.busy = FALSE;          /* board ready */

  printk("%s: address=%#x IRQ=%d.",ADAPTER_ID, BoardData.base, BoardData.irq);
  printk(" 1/15/98 wjasper@tx.ncsu.edu\n");

  return 0;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod das08-AOL" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  if (MOD_IN_USE) {
    printk("%s: device busy, remove delayed.\n", ADAPTER_ID);
    return;
  }

  release_region( BoardData.base, BOARD_SIZE );

#if LINUX_VERSION_CODE > 0x020000
  free_irq(BoardData.irq, NULL);
#else
  free_irq(BoardData.irq);
#endif
  if (unregister_chrdev(MajorNumber, "das08-AOL") != 0) {
    printk("%s: cleanup_module failed.\n", ADAPTER_ID);
  } else {

  #ifdef DEBUG
      printk("%s: module removed.\n", ADAPTER_ID);
  #endif

  }    
}

#ifdef __cplusplus
}
#endif

/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int das08_open(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  int port;

  MOD_INC_USE_COUNT;

  /* 
     check if device is already open: only one process may read from a
     port at a time.  There is still the possibility of two processes 
     reading from two different channels messing things up. However,
     the overhead to check for this may not be worth it.
  */


  if ( minor >= 0 && minor < AD_CHANNELS ) {
    if ( ChanADC[minor].open == TRUE ) {
      return -EBUSY;
    }

    ChanADC[minor].open = TRUE;              /* The device is open */
    ChanADC[minor].gain = UP_10_00V;         /* +/- 10V */
    ChanADC[minor].mode = filePtr->f_flags;  /* set acquisition mode */

    outb_p(0x0, STATUS_REG);

    #ifdef DEBUG
        printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, ChanADC[minor].mode);
    #endif
    return 0;   
  }

  if ( (minor >= AD_CHANNELS) && (minor < AD_CHANNELS+DIO_PORTS) ) {
    port = minor - AD_CHANNELS;
    if ( ChanDIO[port].open == TRUE ) {
      return -EBUSY;
    }

    ChanDIO[port].open = TRUE;                 /* The device is open */
    ChanDIO[port].f_flags = filePtr->f_flags;

    #ifdef DEBUG
      printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, ChanDIO[port].mode);
    #endif
    return 0;
  }

  if ( (minor >= AD_CHANNELS+DIO_PORTS) && (minor < AD_CHANNELS+DIO_PORTS+DA_CHANNELS) ) {
    port = minor - AD_CHANNELS - DIO_PORTS;
    if ( ChanDAC[port].open == TRUE ) {
      return -EBUSY;
    }

    ChanDAC[port].open = TRUE;                 /* The device is open */
    ChanDAC[port].mode = filePtr->f_flags;

    /* zero out the DAC */
    outb(0x0, ChanDAC[port].addr);
    outb(0x0, ChanDAC[port].addr+1);

    #ifdef DEBUG
      printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, ChanDIO[port].mode);
    #endif

  }
  return 0;   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static void das08_close(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  int port;


  MOD_DEC_USE_COUNT;

  if ( minor >= 0 && minor < AD_CHANNELS ) {
    ChanADC[minor].open = FALSE;
  } else if ( minor >= AD_CHANNELS && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    ChanDIO[port].open = FALSE;
  } else if ( minor >= AD_CHANNELS+DIO_PORTS && minor < AD_CHANNELS+DIO_PORTS+DA_CHANNELS ) {
    port = minor - AD_CHANNELS - DIO_PORTS;
    ChanDAC[port].open = FALSE;
  } else {
    printk("das1602_close: Incorrect minor number (%d).\n", minor);
  }  

  #ifdef DEBUG
      printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
  #endif
}


/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int das08_read(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
  int   stat;
  int   port;
  BYTE  bReg;
  int   minor = MINOR(iNode->i_rdev);
  register int i;

  if ( BoardData.busy == TRUE ) {   /* if board is in use, return busy */
      return (-EBUSY);
  } else {
      BoardData.busy = TRUE;
  }

    #ifdef DEBUG
       printk("das1602_read: Entering function\n");
    #endif


  if ( minor >= 0 && minor < AD_CHANNELS ) {

    if ( (ChanADC[minor].count = count) > MAX_COUNT ) {
        printk("das1602_read(): requesting too large a count size.\n");
        BoardData.busy = FALSE;
        return (-1);
    }

    /* Set gains for this channel */
    SetGains((BYTE)ChanADC[minor].gain);

    /* Read */

    switch (ChanADC[minor].mode) {

      case ADC_SOFT_TRIGGER:
          #ifdef DEBUG
              printk("adc_read(): Entering ADC_SOFT_TRIGGER mode.\n");
          #endif
          if (SoftRead(KernBuff, &ChanADC[minor])) {
              printk("das08_read: SoftRead() failed.\n");
              BoardData.busy = FALSE;
              return(-1);
          } 
          break;
     
      case ADC_EXTERNAL_TRIGGER:
          #ifdef DEBUG
              printk("adc_read(): Entering ADC_EXTERNAL_TRIGGER mode.\n");
          #endif
          if (TrigRead(KernBuff, &ChanADC[minor])) {
              printk("das08_read: TrigRead() failed.\n");
              BoardData.busy = FALSE;
              return(-1);
          } 
          break;
    }

    /* Check that data can be written to file. */

    if ((stat = verify_area(VERIFY_WRITE, buf, sizeof(WORD)*ChanADC[minor].count)) != 0) {
      printk("das08_read: Failed VERIFY_WRITE.\n");
      BoardData.busy = FALSE;
      return -1;
    }

    /* Write data to user space */

    for (i = 0; i < ChanADC[minor].count; i++) {
      KernBuff[i] >>= 4;
    }

    if (ChanADC[minor].count == 1) {
      put_fs_word(*KernBuff, (WORD*) buf);
    } else  {
      memcpy_tofs(buf, KernBuff, ChanADC[minor].count*sizeof(WORD));
    }

    BoardData.busy = FALSE;
    return(ChanADC[minor].count);     /* return number of samples read */
  }

  /* check to see if reading a value from the DIO */

  if ( minor >= AD_CHANNELS && minor < AD_CHANNELS + DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    bReg = inb(ChanDIO[port].addr);
    put_fs_byte(bReg, buf);

    #ifdef DEBUG
       printk("das1602_read: %s DIO Port %#x addr %#x set to %#x\n", 
                             ADAPTER_ID, port, ChanDIO[port].addr, bReg);
    #endif
    BoardData.busy = FALSE;
    return 1;
  }

  printk("das08_read: Incorrect minor number (%d).\n", minor);
  BoardData.busy = FALSE;
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int das08_write(struct inode *iNode, struct file *filePtr, const char *buf, int count)
{

  int  minor = MINOR(iNode->i_rdev);
  int  port;
  BYTE bReg;

  /* check to see if writing a value to the DIO */
  if ( (minor >= AD_CHANNELS) && (minor < AD_CHANNELS+DIO_PORTS) ) {
    port = minor - AD_CHANNELS;
    bReg = get_fs_byte(buf);
    ChanDIO[port].value = bReg;
    outb( bReg, ChanDIO[port].addr);

    #ifdef DEBUG
    printk("das08_write %s: DIO Port %#x set to %#x\n", ADAPTER_ID, port, bReg);
    #endif

    BoardData.busy = FALSE;
  } 

  if ( (minor >= AD_CHANNELS+DIO_PORTS) && (minor < AD_CHANNELS+DIO_PORTS+DA_CHANNELS) ) {
    port = minor - AD_CHANNELS - DIO_PORTS;
    ChanDAC[port].value = get_fs_word(buf);
    bReg = get_fs_byte(buf);
    outb( bReg, ChanDAC[port].addr);
    bReg = get_fs_byte(buf+1) & 0xf;
    outb( bReg, ChanDAC[port].addr+1);
    #ifdef DEBUG
    printk("das08_write %s: DAC%d set to %#x  addr %#x\n", 
            ADAPTER_ID, port, ChanDAC[port].value, ChanDAC[port].addr);
    #endif

  }
    return 1;                             /* return number of WORDS written */
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int das08_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  int port;
  int channel;
  BYTE bReg;

  if ( minor >= 0 && minor < AD_CHANNELS ) {
    switch (cmd) {
      case ADC_SET_GAINS:
        ChanADC[minor].gain = (BYTE)arg;
        break;
      case ADC_GET_GAINS:
        put_fs_long((long)ChanADC[minor].gain, (long*) arg);
        break;
      case ADC_COUNTER0:
        LoadCounter0( (LONG) arg );
        break;
      case ADC_COUNTER1:
        LoadCounter1( (LONG) arg );
        break;
      case ADC_COUNTER2:
        LoadCounter2( (LONG) arg );
        break;
      case ADC_GET_STATUS:
        /* return status register */
        put_fs_long( inb_p(STATUS_REG), (long*) arg);
        break;
      case ADC_GET_DIO:
        bReg = inb_p(STATUS_REG) >> 4;
        put_fs_long( bReg, (long*) arg);
        break;
      case ADC_SET_DIO:
        bReg = (BYTE) (arg & 0xf);
        bReg <<= 4;
        BoardData.dio = bReg;
        bReg |=  ChanADC[minor].chan;
        outb_p(bReg, STATUS_REG);
        break;
      default:
        return(-EINVAL);
        break;
    }  /* end switch */
  }  /* end if */

  if ( minor >= AD_CHANNELS && minor < AD_CHANNELS+DIO_PORTS ) {
    port = minor - AD_CHANNELS;
    switch (cmd) {
      case DIO_SET_MODE:
        arg &= 0x3;
        switch (port) {
	    case 0:		       /* Port A */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port A\n");
              #endif
              dio_reg &= 0x9f;
              dio_reg |= (arg << 5);
              outb_p(dio_reg, DIO_CNTRL_REG);
              ChanDIO[port].mode = arg;
              break;

            case 1:			/* Port 1B */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port 1B\n");
              #endif
              if (arg & 0x2)
	        return(-EINVAL);	       /* Port 1B only has Modes 0 & 1 */
              dio_reg &=  0xfb;
              dio_reg |= (arg << 2);
              outb_p(dio_reg, DIO_CNTRL_REG);
              ChanDIO[port].mode = arg;
              break;

            case 2:			/* Port 1C */
              #ifdef DEBUG
              printk("DIO_SET_MODE for Port 1C\n");
              #endif
              if (arg)
	        return(-EINVAL);	/* Port 1C only has Mode 0 I/O */
              ChanDIO[port].mode = arg;
              break;

            default:
              #ifdef DEBUG
              printk("DIO_SET_MODE for Invalid Port\n");
              #endif
              return(-EINVAL);	/* Wrong Port Number */
              break;
        }
        break;
    
      case DIO_SET_DIRECTION:

        #ifdef DEBUG
        printk("DIO_SET_DIRECTION: minor = %d, arg = %d\n", (int) minor, (int) arg );
        #endif

        switch (port) {
          case 0:			/* Port A */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port A\n");
            #endif
            arg &= 0x1;
            dio_reg &=  0xef;
            dio_reg |= (arg << 4);
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          case 1:			/* Port B */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port B\n");
            #endif
            arg &= 0x1;
            dio_reg &=  0xfd;
            dio_reg |= (arg << 1);
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          case 2:			/* Port C */
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Port C\n");
            #endif
            switch ( arg ) {
              case PORT_INPUT:         /* CU = IN  CL = IN */
                  dio_reg |= 0x9;
                  break;
              case PORT_OUTPUT:        /* CU = OUT  CL = OUT */
                  dio_reg &= ~(0x9);
                  break;
              case LOW_PORT_INPUT:     /* CL = IN */
                  dio_reg |= 0x1;
                  break;
              case LOW_PORT_OUTPUT:    /* CL = OUT */
                  dio_reg &= ~(0x1);
                  break;
              case HIGH_PORT_INPUT:    /* CU = IN */
                  dio_reg |= 0x8;
                  break;
              case HIGH_PORT_OUTPUT:   /* CU = OUT */
                  dio_reg &= ~(0x8);
                  break;
              default:
                  return(-EINVAL);
                  break;
            }
            outb_p(dio_reg, DIO_CNTRL_REG);
            break;

          default:
            #ifdef DEBUG
            printk("DIO_SET_DIRECTION for Invalid Port\n");
            #endif
            return(-EINVAL);
            break;
        }
        break;
    
      default:
        return(-EINVAL);
        break;
    }  /* end switch */
  }    /* end if */


  if ( minor >= AD_CHANNELS+DIO_PORTS && minor < AD_CHANNELS+DIO_PORTS+DA_CHANNELS ) {
    channel = minor - AD_CHANNELS - DIO_PORTS;

  }
  return 0;
}

/***************************************************************************
 *
 *  Alarm handler to handle situations where an interrupt is missed.
 *
 ***************************************************************************/

static void das08_TimerHandler(LONG unused)
{
  /* 
     we should only be here if an interrupt occured while 
     we were reading half the FIFO 
  */

  #ifdef DEBUG
      printk("TimerHandler: WordsToRead = %d\n", WordsToRead);
  #endif

  wake_up_interruptible(&adc0_wait);  
}

/***************************************************************************
 *
 * Interrupt handler used to service enhanced mode(interrupt) read().
 *
 ***************************************************************************/

static void das08_ReadInterrupt(int irq)
{
  WORD msb, lsb;

  #ifdef DEBUG
      printk("Entering das08_ReadInterrupt(). irq = %d mode = %d  count = %d\n", 
              irq, CurrChan->mode, WordsToRead );
  #endif

  cli();
  if ( WordsToRead == 0 ) return;

  outb_p(0x1, MSB_DATA_BYTE);                 /* Force conversion */
  del_timer(&TimerList);

  /* this board is so slow that you need to set the MUX before the conversion is done! */
  outb_p(CurrChan->chan | INTE | BoardData.dio, STATUS_REG);  /* Set MUX and Interrupt */

  /* poll until done (Yech!) */
  while (inb_p(STATUS_REG) & EOC);

  lsb = (WORD) inb(LSB_AND_CHNLS);            /* Sample: ADC -> kernel buffer */
  msb = (WORD) inb(MSB_DATA_BYTE);

  *KernBuffPtr++ = (msb << 8) | lsb;
  WordsToRead--;                              /* One sample at a time */

  if (!WordsToRead) {
    wake_up_interruptible(&adc0_wait);
  } else {
     TimerList.expires =  jiffies + 600;
     add_timer(&TimerList);
  }
  sti();
}

/***************************************************************************
 *
 * Handles software read().
 *
 * Bang! Bang! Method: 
 *    
 *    o Force interrupt by forcing conversion.
 *    o Get one sample at a time and put it in the kernel buffer(kernBuff).
 *    o Get chan->count samples
 *
 ***************************************************************************/

static int SoftRead(WORD *kernBuff, ADC_ChanRec *chan)
{
  WORD msb, lsb;

  #ifdef DEBUG
      printk("Entering SoftRead().\n");
  #endif

  /* Prepare global data for parameterless interrupt handler */

  KernBuffPtr = KernBuff;                      /* same with KernBuff           */
  WordsToRead = chan->count;

  outb_p(chan->chan | BoardData.dio, STATUS_REG);  /* Set MUX channel */

  while ( WordsToRead ) {
    outb_p(0x1, MSB_DATA_BYTE);                /* Force conversion */

    /* poll until done (Yech!) */
    while (inb_p(STATUS_REG) & EOC);

    lsb = (WORD) inb(LSB_AND_CHNLS);           /* Sample: ADC -> kernel buffer */
    msb = (WORD) inb(MSB_DATA_BYTE);
    *KernBuffPtr++ = (msb << 8) | lsb;
    WordsToRead--;                              /* One sample at a time */
  }

  return 0;
}

static int TrigRead(WORD *kernBuff, ADC_ChanRec *chan)
{

  #ifdef DEBUG
      printk("Entering TrigRead().\n");
  #endif

  /* Prepare global data for parameterless interrupt handler */
  CurrChan = chan;                         /* pass chanel number to global */
  KernBuffPtr = KernBuff;                  /* same with KernBuff           */
  WordsToRead = chan->count;

  outb_p(chan->chan | INTE | BoardData.dio, STATUS_REG);    /* Set MUX and Interrupt */
  TimerList.expires =  jiffies + 600;                   /* just in case ...      */
  add_timer(&TimerList);

  interruptible_sleep_on(&adc0_wait);             /* Block in wait state */

  if ( WordsToRead != 0 ) {
      printk("Timing error in TrigRead: WordsToRead = %d\n", WordsToRead);
      return -1;
  }
  #ifdef DEBUG
    printk("Leaving TrigRead.\n");
  #endif

  return 0;
}

/***************************************************************************
 *
 * Set voltage gains.
 *
 ***************************************************************************/

static void SetGains( BYTE gain )
{
  #ifdef DEBUG
      printk("Entering SetGains(). Setting gain to %d\n", gain);
  #endif

  outb_p(gain, GAIN_REG);
}

/***************************************************************************
 *
 * Load value into Counter 0  XXXX    Mode    MSB     LSB
 *                            Byte 3  Byte 2  Byte 1  Byte 0
 *
 ***************************************************************************/

static void LoadCounter0( LONG value )
{
  BYTE mask;
  BYTE bData;

  /* Write the value into Counter 0 Mode 2 */

  #ifdef DEBUG
      printk("LoadCounter0: load value %#lx into Counter 0.\n", value);
  #endif

  /* the mode is in the thrid byte */
  mask = 0xff & (value >> 16);
  mask += C0+LSBFIRST;
  outb_p(mask, COUNTER_CONTROL); 

  if ( value & 0xff000000 ) {       /* load control word only */
      return;
  } else {
      /* LSB in the first byte of value */
      bData = (BYTE) (value & 0xff);
      outb_p(bData, COUNTER_0_DATA); 

      /* MSB in the second byte of value */
      bData = (BYTE) ((value >> 8) & 0xff);
      outb_p(bData, COUNTER_0_DATA); 
  }
}

static void LoadCounter1( LONG value )
{
  BYTE mask;
  BYTE bData;

  /* Write the value into Counter 1 Mode 2 */

  #ifdef DEBUG
      printk("LoadCounter0: load value %#lx into Counter 0.\n", value);
  #endif

  /* the mode is in the thrid byte */
  mask = 0xff & (value >> 16);
  mask += C1+LSBFIRST;
  outb_p(mask, COUNTER_CONTROL); 

  if ( value & 0xff000000 ) {       /* load control word only */
      return;
  } else {
      /* LSB in the first byte of value */
      bData = (BYTE) (value & 0xff);
      outb_p(bData, COUNTER_1_DATA); 

      /* MSB in the second byte of value */
      bData = (BYTE) ((value >> 8) & 0xff);
      outb_p(bData, COUNTER_1_DATA); 
  }
}

static void LoadCounter2( LONG value )
{
  BYTE mask;
  BYTE bData;

  /* Write the value into Counter 2 Mode 2 */

  #ifdef DEBUG
      printk("LoadCounter0: load value %#lx into Counter 0.\n", value);
  #endif

  /* the mode is in the thrid byte */
  mask = 0xff & (value >> 16);
  mask += C2+LSBFIRST;
  outb_p(mask, COUNTER_CONTROL); 

  if ( value & 0xff000000 ) {       /* load control word only */
      return;
  } else {
      /* LSB in the first byte of value */
      bData = (BYTE) (value & 0xff);
      outb_p(bData, COUNTER_2_DATA); 

      /* MSB in the second byte of value */
      bData = (BYTE) ((value >> 8) & 0xff);
      outb_p(bData, COUNTER_2_DATA); 
  }
}


